from enum import Enum
import operator

class AST(object):
    def eval(self):
        raise NotImplementedError()

""" Base node types """

class Statement(AST):
    def __init__(self):
        super(Statement,self).__init__()

class Expression(AST):
    def __init__(self):
        super(Expression,self).__init__()

""" Expressions """
class BinOp(Expression):
    def __init__(self, left,right):
        super(Expression,self).__init__()
        self.left = left
        self.right = right

class AddOp(BinOp):
    def eval(self):
        return self.left.eval() + self.right.eval()

class SubOp(BinOp):
    def eval(self):
        return self.left.eval() - self.right.eval()

class MultOp(BinOp):
    def eval(self):
        return self.left.eval() * self.right.eval()

class DivOp(BinOp):
    def eval(self):
        left  = self.left.eval()
        right = self.right.eval()

        if right == 0:
            raise runtime.Errors.ZeroDivisionError()

        return left / right

class ModOp(BinOp):
    def eval(self):
        left  = self.left.eval()
        right = self.right.eval()

        if right == 0:
            raise runtime.Errors.ZeroDivisionError()

        return left % right

"""
# Unary arithmetic operations
"""
class UnaryOp(Expression):
    def __init__(self, op, operand):#Expression):
        super(UnaryOp,self).__init__()
        self.op = op
        self.operand = operand

    def eval(self):
        if self.op == '+':
            return self.operand.eval()
        elif self.op == '-':
            return -(self.operand.eval())
        else:
            raise ValueError('Unsupported unary operation!')


"""
# Base class for comparisons.
"""
class Compare(Expression):

    class Op(Enum):
        AND = 1
        OR  = 2
        NOT = 3

    opTable = {
        '<'  : operator.lt,
        '>'  : operator.gt,
        '==' : operator.eq,
        '>=' : operator.ge,
        '<=' : operator.le,
        '!=' : operator.ne,
        Op.AND : operator.__and__,
        Op.OR  : operator.__or__,
        Op.NOT : operator.__not__,
    }

    def __init__(self, op):
        super(Op,self).__init__()
        self.op = op

class BinaryComp(Compare):
    def __init__(self, left, right, op):
        super(BinaryComp).__init__(op=op)
        self.left = left
        self.right = right

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()

        return Compare.opTable[self.op](left, right)

class UnaryComp(Compare):
    def __init__(self, operand, op):
        super(UnaryComp,self).__init__(op=op)
        self.operand = operand

    def eval(self):
        operand = self.operand.eval()
        return Compare.opTable[self.op](operand)


"""
# Represents None, False and True literals.
"""
class NameConstant(Expression):
    nameTable = { 'None' : None, 'True': True, 'False': False }

    def __init__(self, name):
        super(NameConstant,self).__init__()
        self.name = name

    def eval(self):
        try:
            return NameConstant.nameTable[self.name]
        except KeyError:
            raise ValueError("Wrong name constant")

"""
# A variable name.
#     @id holds the name as a string
"""
class Name(Expression):

    def __init__(self):
        super(Name,self).__init__()
        self.id = id

    def eval(self):
        return self.id
    
"""
# Number literal
"""
class Num(Expression):

    def __init__(self, value):
        super(Num,self).__init__()
        self.value = value

    def eval(self):
        return self.value

"""
# String literal
"""
class Str(Expression):
    def __init__(self, value):
        super(Str,self).__init__()
        self.value = value

    def eval(self):
        return self.value
